unit unit3;

{$mode objfpc}{$H+}

interface

uses
   Classes, SysUtils, Forms, Controls, Graphics, StdCtrls, Dialogs,
   ExtCtrls,LCLtype, decoupe, unit2;

const
   ln10 = ln(10);

   nf_ini = 'CCD_GRAPPA.ini';
   nf_out = 'CCD_Grappa.out';
type


   { TForm1 }

   TForm1 = class(TForm)
      ButtonAide: TButton;
      ButtonAstroPhoto: TButton;
      ButtonAnalyser: TButton;
      ButtonBase: TButton;
      CheckBoxCollapse: TCheckBox;
      CheckBoxXMax: TCheckBox;
      CheckBoxDebog: TCheckBox;
      CheckBoxMagLimite: TCheckBox;
      CheckBoxErreur: TCheckBox;
      ComboBoxFiltre: TComboBox;
      EditMagLimite: TEdit;
      EditErreur: TEdit;
      GroupBox1: TGroupBox;
      Label1: TLabel;
      LabeledEditDonnees: TLabeledEdit;
      LabeledEditBase: TLabeledEdit;
      MemoOut: TMemo;
      OpenDialog1: TOpenDialog;
      RadioButtonSIRIL: TRadioButton;
      RadioButtonCSVFlux: TRadioButton;
      RadioButtonPRISM11: TRadioButton;
      RadioButtonCSVFluxMag: TRadioButton;
      SaveDialog1: TSaveDialog;
      SelectDirectoryDialog1: TSelectDirectoryDialog;
      procedure ButtonAideClick(Sender: TObject);
      procedure ButtonBaseClick(Sender: TObject);
      procedure ButtonAnalyserClick(Sender: TObject);
      procedure ButtonAstroPhotoClick(Sender: TObject);
      procedure CheckBoxCollapseChange(Sender: TObject);
      procedure CheckBoxDebogChange(Sender: TObject);
      procedure CheckBoxXMaxChange(Sender: TObject);
      procedure CheckBoxMagLimiteChange(Sender: TObject);
      procedure CheckBoxErreurChange(Sender: TObject);
      procedure EditErreurChange(Sender: TObject);
      procedure EditMagLimiteChange(Sender: TObject);
      procedure FormCreate(Sender: TObject);
      procedure FormDestroy(Sender: TObject);
      procedure LabeledEditBaseChange(Sender: TObject);
      procedure RadioButtonPRISM11Change(Sender: TObject);



   private

   public
   end;

type
   TGRAPPA = bitpacked record
                                  //         Size
                                  // Unit   (byte) Definition
             Source_ID : QWord;   //          8    Unique within a particular Data Release. See https://dms.cosmos.esa.int/COSMOS/doc_fetch.php?id=2779219
             RA        : int32;   // mas      4    R.A.truncated to 1mas precision truncation: +10.123 = 10
             RA2       : byte;    // 4µas     1    R.A.component <1mas. Units of 4 µas. To be added to RA. 30 = 0.120mas
             e_RA      : uint16;  // 10µas    2    Uncertainty in RA units of 10 µas
             DEC       : int32;   // mas      4    Declination+90° truncated 1 mas precision truncation: +10.123 = 10
             DEC2      : byte;    // 4µas     1    DEC component <1mas. Units of 4 µas. To be added to DEC. 30 = 0.120mas
             e_DEC     : uint16;  // 10µas    2    Uncertainty in Dec, units of 10 µas
             parallax  : uint16;  // 12.5µas  2    Permet de coder la parallax de Proxima Cen.
             e_parallax: uint16;  // 10µas    2    Uncertainty in Parallax units of 10 µas
             PM_RA     : int32;   // µas/yr   4    Annual proper motion in RA. Value has been multiplied by Cos(declination)
             e_PM_RA   : uint16;  // µas/yr   2    Uncertainty in Proper motion in RA
             PM_DEC    : int32;   // µas/yr   4    Annual proper motion in Dec
             e_PM_DEC  : uint16;  // µas/yr   2    Uncertainty in Proper motion in Dec
             RUWE      : byte;    //   0.1    1    Value for ruwe indicator See : https://dms.cosmos.esa.int/COSMOS/doc_fetch.php?id=3757412
                                  //               RUWE is set to 0 for sources with only two parameters (RA & DEC)
             has_RV,              //         1/8   Set if there is a radial velocity for the star in the RV.dat file
             has_PM_parallax_RUWE,//         1/8   Set if there are PM, parallax and RUWE values for the star
             has_hipparcos2_id,   //         1/8   Set if an Hipparcos2 crossmatch has been found
             has_TYCHO2DSC_id,    //         1/8   Set if an Tycho2DSC crossmatch has been found
             has_G,has_BP,has_RP, //       3x1/8   Set if magG, magBP, magRP are given respectively
             isDuplicatedSource : boolean;// 1/8   Set if this source happened to been duplicated and one source only has been kept
             G,BP,RP   : int32;   // 0.1mmag 3x4   Values for magnitude in the G, BP and RP bands respectively including e_G, e_BP, e_RP
                                  //               Coded as : 18bits for value (value can go up to 26.2144), 14 bits for incertitudes.
          end;

         TSource = record
            Source_ID : QWord;
            RA,DEC    : double;
            magG,magG_BP,magG_RP : real;
         end;

         TZone  = record
            nb_source : integer;
            Liste_source : array of TSource;
         end;

         TObjetCCD = record
            X,Y          : real;
            RAhms,DEChms : string;
            RA,DEC,FwhmX,FwhmY,Angle,
            Mag,Mag_calc,Flux,OMC,SNR,
            Constante                 : double;
            trouve,EstEtalon          : boolean;
            zone                      : TZone;
         end;

         TFiltre = (G,G_BP,G_RP,gp,rp,ip);
         TOrigine = (PRISMV11,SIRIL,CSVFluxMag,CSVFlux);

const
   nom_des_filtres : array[TFiltre] of string = ('G','G_BP','G_RP','gp','rp','ip');

var
   Form1: TForm1;

   nf_astro, nf_photo,base : string;
   ok_astro, ok_photo,
   ok_collapse, ok_xmax, ok_erreur  : boolean;
   debog : boolean;
   fd    : textfile;

implementation

{$R *.lfm}


function power10(a : double):double;
begin
  power10:=exp(a*ln10);
end;

function log(a : double):double;
begin
  log:=ln(a)/ln(10)
end;

Procedure Rechercher_tout(NFGrappa : string; RAMin,RAMax,DecMin,DECMax,Maglimite : double; filtre : TFiltre; var zone : TZone; var fd : textfile);
{ Procédure de recherche dans un fichier grappa des spources comrise dans une zone du ciel
Crée la liste_source dans le record zone }
var
  fGRAPPA    : file of TGRAPPA;
  LObjet     : TGRAPPA;
  i,lFGrappa : longint;
  magnitude,
  vRA,vDEC   : double;

begin

  if fileExists(NFGrappa) then
  begin
     AssignFile(fGRAPPA,NFGrappa);
     Reset(fGRAPPA);

     lfGRAPPA:=Filesize(fGRAPPA)-1;
     if debog then Writeln(fd,DateTimeToStr(now),' Ouverture du fichier ',NFGrappa,'. Longueur :',lfGRAPPA,' enregistrments.');
     if debog then Writeln(fd,DateTimeToStr(now),' Recherche avec limite de magnitude ',nom_des_filtres[filtre],' : ',MagLimite:6:3);
     i:=-1;
     repeat
        inc(i);
        Seek(fGRAPPA,i);
        Read(fGRAPPA,LObjet);
        with LObjet do
        begin
           vRA:=(RA+RA2/250)/3600000;
           vDEC:=(DEC+DEC2/250)/3600000-90;
        end;

        // version du code valable pour un test sur G

        magnitude:=100;
        case Filtre of
        G    : if LObjet.has_G  then magnitude:=(LObjet.G  AND $3FFFF)/10000;
        G_BP : if LObjet.has_BP then magnitude:=(LObjet.BP AND $3FFFF)/10000;
        G_RP : if LObjet.has_RP then magnitude:=(LObjet.RP AND $3FFFF)/10000;
        end;

        if (vRa>=RAmin) and (vRA<=RAMax) and (vDEC<=DECMax) and (vDEC>=DECmin) and (magnitude<=MagLimite) then
        begin
           Inc(zone.nb_source);
           SetLength(zone.liste_source,zone.nb_source+1);

              with zone.Liste_source[zone.nb_source] do
              begin
                 Source_ID:=LObjet.Source_ID;
                 RA:=vRA;
                 DEC:=vDEC;
                 magG:=99; magG_BP:=99; magG_RP:=99;  { pour sommer des flux nuls en cas d'utilisation }
                 if LObjet.has_G  then magG:=   (LObjet.G  AND $3FFFF)/10000;
                 if LObjet.has_BP then magG_BP:=(LObjet.BP AND $3FFFF)/10000 ;
                 if LObjet.has_RP then magG_RP:=(LObjet.RP AND $3FFFF)/10000 ;
              end;
        end;
     until (i=lfGRAPPA);
     if debog then Writeln(fd,DateTimeToStr(now),' Nombre de source_ID détectés : ',zone.nb_source);
     Closefile(fGRAPPA);
  end
  else
     if debog then Writeln(fd,DateTimeToStr(now),' Fichier ',NFGrappa,' inexistant');

end;

procedure Explore_Zone(RAMin,RAMax,DecMin,DECMax,Maglimite : double; Base : string; filtre : TFiltre; var zone : TZone; var fd : Textfile);
{ effectue un parcours des fichiers GRAPPA couvrant la zone du ciel étudiée }
  function Max(A,B : double):double;
  begin
     if A>B then Max:=A else Max:=B
  end;
  function Min(A,B : double):double;
  begin
     if A<B then Min:=A else Min:=B
  end;

var
  i,j,iRAmin,iRAmax,iDECmin,iDECmax : integer;
  NordFait,SudFait : boolean;
  sRA,sDEC,fichier : string;

begin
  // on commence par effacer la zone

  with zone do nb_source:=0;
  If Debog then Writeln(fd,DateTimeToStr(now),' Création de la zone de recherche');
  If Debog then Writeln(fd,DateTimeToStr(now),' Explore_zone : exploration de GRAPPA sur la zone RA min : ',RAMin:12:7,
                         ' RA Max : ',RAMax:12:7,' - DEC min :',DecMin:12:7,
                         ' DEC Max : ',DECMax:12:7,' avec limite de magnitude ',nom_des_filtres[filtre],' : ',MagLimite:6:3);

  NordFait:=false;
  Sudfait :=false;

  iRAmin:=Trunc(Max(Min(RAmin,360),0));
  iRAmax:=Trunc(Max(Min(RAmax,360),0));
  iDECmin:=Trunc(90+DecMin);  // iDEC = indice <> Déclinaison
  iDECmax:=Trunc(90+DecMax);  // iDEC = indice <> Déclinaison

  for i:=iRAmin to iRAmax do
     for j:=iDECmin to iDECmax do
     begin
        if iDECMax>174 then
        begin
           if not NordFait then begin
              Fichier:=Base+'\N85-90';
              Rechercher_tout(Fichier,RAMin,RAMax,DecMin,DECMax,Maglimite,filtre,zone,fd);
              Nordfait:=true;
              If Debog then Writeln(fd,DateTimeToStr(now),' Analyse faite sur le pôle Nord');
           end;
        end else
        begin
           if iDECMax<5 then
           begin
              if not SudFait then begin
                 Fichier:=Base+'\S85-90';
                 Rechercher_tout(Fichier,RAMin,RAMax,DecMin,DECMax,Maglimite,filtre,zone,fd);
                 Sudfait:=true;
                 If Debog then Writeln(fd,DateTimeToStr(now),' Analyse faite sur le pôle Sud');
              end;
           end
           else   // cas de la zone principale non polaire
           begin
              Str(i,sRA);
              Str(j,sDEC);
              fichier:=Base+'\'+sDEC+'\'+sRA+'-'+sDEC;
              Rechercher_tout(Fichier,RAMin,RAMax,DecMin,DECMax,Maglimite,filtre,zone,fd);
              If Debog then Writeln(fd,DateTimeToStr(now),' Analyse faite sur le fichier ',fichier);
           end;
        end;
     end;
end;


{ TForm1 }
procedure TForm1.CheckBoxCollapseChange(Sender: TObject);
begin
  if CheckBoxCollapse.checked and CheckBoxXMax.Checked then CheckBoxXMax.checked:=false;;
end;

procedure TForm1.CheckBoxDebogChange(Sender: TObject);
begin

end;

procedure TForm1.ButtonAstroPhotoClick(Sender: TObject);
begin
   if opendialog1.Execute then
   begin
      if FileExists(opendialog1.FileName) then
      begin
         ok_astro:=true;
         nf_astro:=opendialog1.FileName;
         LabeledEditDonnees.Text:=nf_astro;
      end
      else
      begin
         ok_astro:=false;
         nf_astro:='';
      end;
      ButtonAnalyser.Enabled:=ok_astro;
   end;
end;

function Deg2DMS(angle : real):string;
var
   ad,am      : integer;
   reste,asec : real;
   sd,sm,ss   : string;
begin
   ad :=trunc(angle);
   reste:=Abs(angle-ad)*60;
   am :=trunc(reste);
   asec:=(reste-am)*60;
   Str(Abs(ad):2,sd); sd:=trim(sd); if ad<10 then sd:='0'+sd; if angle<0 then sd:='-'+sd else sd:='+'+sd;
   Str(am:2,sm);     sm:=trim(sm); if am<10 then sm:='0'+sm;
   Str(asec:5:3,ss); ss:=trim(ss); if asec<10 then ss:='0'+ss;
   Deg2DMS:=sd+'d'+sm+''''+ss+'''''';
end;

function Deg2HMS(angle : real):string;
var
   ad,am      : integer;
   reste,asec : real;
   sd,sm,ss   : string;
begin
   angle:=angle/15;
   ad :=trunc(angle);
   reste:=Abs(angle-ad)*60;
   am :=trunc(reste);
   asec:=(reste-am)*60;
   Str(Abs(ad):2,sd); sd:=trim(sd); if ad<10 then sd:='0'+sd; if angle<0 then sd:='-'+sd;
   Str(am:2,sm);     sm:=trim(sm); if am<10 then sm:='0'+sm;
   Str(asec:5:3,ss); ss:=trim(ss); if asec<10 then ss:='0'+ss;
   Deg2HMS:=sd+'h'+sm+'m'+ss+'s';
end;

procedure TForm1.ButtonAnalyserClick(Sender: TObject);
var
   buffer,mot                       : String;
   doCollapse,doBrightest,
   doSauve                          : boolean;
   f_astro,fo                       : textfile;
   nb_etoile,i,j,k,cpt_non_trouve,
   cpt_corr_mult,cpt_trouve,
   cpt_Etalon_supprime,
   Nb_source,iMax,n                 : longint;
   code,nb_mot,scode,cpt_passe      : integer;
   PhotoObjet                       : TObjetCCD;
   Table                            : Array of TObjetCCD;
   RAMin,RAMax,DECMin,DECMax,
   magMaxBright,MagLimite,FWHMmax,
   magCollapseG,magCollapseG_BP,
   magCollapseG_RP,MagTest,MagEtalon,
   magEtalon_1,magEtalon_i,FluxTheorique,
   FluxEtalon,SommeSigmaMesure,
   sigmaMesure,SommeSigmaPrism,
   SommeConstante,SigmaConstante,
   ConstanteMoyenne,SigmaPrism      : double;
   CCDZone                          : TZone;
   filtre                           : TFiltre;
   mot_csv                          : T_phrase;
   Origine                          : TOrigine;
begin
   ButtonAnalyser.Caption:='Attendez...';


   if CheckBoxErreur.Checked then begin
      Val(EditErreur.Text,FWHMmax,code);
      if (code<>0) or (FWHMMax<=0) then FWHMmax:=1;
   end
      else FWHMMax:=1;
   FWHMMax:=FWHMMax/3600; { 1" exprimée en ° }

   if CheckBoxMagLimite.Checked then begin
      Val(EditMagLimite.Text,Maglimite,code);
      if code<>0 then
         MagLimite:=99;
   end
   else Maglimite:=99;

   debog:=CheckBoxDebog.Checked;
   if debog then begin
      AssignFile(fd,'AnalyseCCD-GRAPPA.log');
      Rewrite(fd);
   end;

   if CheckboxCollapse.Checked then doCollapse:=true;
   if CheckBoxXMax.Checked then doBrightest:=true;
   if RadioButtonPRISM11.Checked then Origine:=PRISMV11;
   if RadioButtonSIRIL.Checked then Origine:=SIRIL;
   if RadioButtonCSVFluxMag.Checked then Origine:=CSVFluxMag;
   if RadioButtonCSVFlux.Checked then Origine:=CSVFlux;


   if debog then begin
      Writeln(fd,DateTimeToStr(now),' CCD-GRAPPA v1.2 Mode "Bavard" actif');
      Writeln(fd,DateTimeToStr(now),' Fichier de données à analyser : ',LabeledEditDonnees.Text);
      if doCollapse then Writeln(fd,DateTimeToStr(now),' Mode "Cumul photometrique" actif.');
      if doBrightest then Writeln(fd,DateTimeToStr(now),' Mode "Objet le plus brillant" actif.');
      Writeln(fd,DateTimeToStr(now),' Erreur de correspondance admise en RA et DEC : ',FWHMMax*3600:5:3,' arcsec.');
      Writeln(fd,DateTimeToStr(now),' Magnitude limite de recherche dans GRAPPA : ',Maglimite:6:3);

      Case Origine of
      PRISMV11   : Writeln(fd,DateTimeToStr(now),' Données issues de PRISM V11');
      CSVFluxMag : Writeln(fd,DateTimeToStr(now),' Données issues d''un fichier CSV avec flux et magnitude');
      CSVFlux    : Writeln(fd,DateTimeToStr(now),' Données issues d''un fichier CSV avec flux');
      end;

   end;


   AssignFile(f_astro,LabeledEditDonnees.Text);
   Reset(f_astro);

   RAmin:=360; DECmin:=90;
   RAMax:=0;   DECMax:=-90;

   Case Origine of
   PRISMV11 :
   begin
      Readln(f_astro);   // on saute une ligne d'entete
      Readln(f_astro,nb_etoile);
      Readln(f_astro);   // on saute une ligne

      if debog then Writeln(fd,DateTimeToStr(now),' Le fichier astro contient les donnees de ',nb_etoile, ' objets.');

      // On lit toutes les donnees astro :

      SetLength(Table,nb_etoile+1);

      if debog then Writeln(fd,DateTimeToStr(now),' Memoire reservee pour ',nb_etoile,' objets.');


      code:=0;
      for i:=1 to nb_etoile do      // lecture de tout le fichier Astrometrie
      begin
         With Table[i] do
         begin
            buffer:='';

            Readln(f_astro,buffer);

            mot:=Copy(buffer,7,10); mot:=trim(mot);      // Mot 1 : X
            Val(mot,X,code);
            mot:=Copy(buffer,17,10); mot:=trim(mot);     // Mot 2 : Y
            Val(mot,Y,code);
            RAhms:=Copy(buffer,28,13); mot:=trim(mot);   // Mot 3 : RAhms
//            DEChms:=Copy(buffer,44,14); mot:=trim(mot);  // Mot 4 : DEChms
//            DEChms[4]:='d'; // ReplaceStr(DEChms,'°','d');
            mot:=Copy(buffer,60,12); mot:=trim(mot);     // Mot 5 : RA
            Val(mot,RA,code);
            mot:=Copy(buffer,73,12); mot:=trim(mot);     // Mot 6 : DEC
            Val(mot,DEC,code);
            DEChms:=Deg2DMS(DEC);
            if RA<RAmin then RAmin:=RA;
            if RA>RAMax then RAMax:=RA;
            if DEC<DECmin then DECmin:=DEC;
            if DEC>DECMax then DECMax:=DEC;

            FwhmX:=0;
            FwhmY:=0;
            Angle:=0;
            Mag:=0;
            Flux:=0;
            OMC:=0;
            SNR:=0;
         end;
      end;


      // la zone du ciel où devra s'effectuer la recherche est déterminée
      //
      if debog then Writeln(fd,DateTimeToStr(now),' Fin de lecture de la partie astrometrie du fichier.');
   end
   else
   begin                    // fichier SIRIL ou csv
      nb_etoile:=0;
      if Origine=SIRIL then  Readln(f_astro);   // on saute une ligne d'entete
      Repeat
         readln(f_astro,buffer);
         decouper(buffer,',',mot_csv,nb_mot);


         scode:=0;
         with PhotoObjet do
         begin
            if Origine=SIRIL then
            begin
               val(mot_csv[6],RA,code);   scode:=scode+code;
               val(mot_csv[7],DEC,code);  scode:=scode+code;
               val(mot_csv[10],Mag,code); scode:=scode+code;
               Flux:=Power10(Mag/2.5);
            end else
            begin
               val(mot_csv[1],RA,code);   scode:=scode+code;
               val(mot_csv[2],DEC,code);  scode:=scode+code;
               val(mot_csv[3],Flux,code); scode:=scode+code;
               mag:=99;
               if Origine=CSVFluxMag then
                  val(mot_csv[4],mag,code);
            end;
         end;

         if scode=0 then  // on a tout bien lu de la ligne du .csv
         begin
            Inc(nb_etoile);
            SetLength(Table,nb_etoile+1);
            with Table[nb_etoile] do
            begin
               RA:=PhotoObjet.RA;      RAhms:='             ';
               DEC:=PhotoObjet.DEC;    DEChms:='              ';
               Flux:=PhotoObjet.Flux;
               mag:=PhotoObjet.mag;
               if RA<RAmin then RAmin:=RA;
               if RA>RAMax then RAMax:=RA;
               if DEC<DECmin then DECmin:=DEC;
               if DEC>DECMax then DECMax:=DEC;

            end;
         end;

      until eof(f_astro);

      if debog then
         if Origine=SIRIL then Writeln(fd,DateTimeToStr(now),' Fin de lecture du fichier d''origine SIRIL.')
         else
            Writeln(fd,DateTimeToStr(now),' Fin de lecture du fichier .csv.');

      if debog then Writeln(fd,DateTimeToStr(now),' Memoire reservee pour ',nb_etoile,' objets.');
   end;
   end;

   Filtre:=G;
   case ComboBoxFiltre.ItemIndex of
      0 : Filtre:=G;
      1 : Filtre:=G_BP;
      2 : Filtre:=G_RP;
      else  Filtre:=G;
      end;
   if CheckBoxMagLimite.Checked then
   begin
      if debog then
         Writeln(fd,DateTimeToStr(now),' ComboBoxFiltre.ItemIndex = ',ComboBoxFiltre.ItemIndex,'. Filtrage : sources de magnitude ',
                    nom_des_filtres[filtre],' < ',Maglimite:6:3);
   end
      else
         if debog then Writeln(fd,DateTimeToStr(now),' Pas de filtrage en magnitude');


   With CCDZone do nb_source:=0;

   RAmin:=RAmin-0.0002;    // 1/3 arcsec soit une zone de recherche de 2x2 arcsec
   RAMax:=RAMax+0.0002;
   DECmin:=DECmin-0.0002;
   DECMax:=DECMax+0.0002;

   if debog then Writeln(fd,DateTimeToStr(now),' Exploration de GRAPPA sur la zone RA min : ',RAMin:12:7,
                         ' RA Max : ',RAMax:12:7,' - DEC min :',DecMin:12:7,
                         ' DEC Max : ',DECMax:12:7,' Magnitude limite : ',maglimite:6:3);

   // on recherche toutes les sources GRAPPA définie dans cette zone :

   Explore_Zone(RAMin,RAMax,DecMin,DECMax,MagLimite,Base,Filtre,CCDZone,fd);

   if debog then Writeln(fd,DateTimeToStr(now),' La zone CCD contient ',CCDZOne.nb_source,' source_id dans le catalogue.');

   AssignFile(fo,nf_out);
   Rewrite(fo);

   if CCDZOne.nb_source>0 then
   begin

      if Origine=PRISMV11 then begin

         Readln(f_astro);   // on saute une ligne
         Readln(f_astro);   // on saute une ligne
         Readln(f_astro);

      end;

      cpt_non_trouve:=0;

      for i:=1 to nb_etoile do
      begin
         if Origine=PRISMV11 then With PhotoObjet do
         begin
            buffer:='';


            Readln(f_astro,buffer);

            mot:=Copy(buffer,8,8); mot:=trim(mot);   // Mot 1 : X
            Val(mot,PhotoObjet.X,code);
            mot:=Copy(buffer,17,8); mot:=trim(mot);   // Mot 2 : Y
            Val(mot,PhotoObjet.Y,code);

            j:=0; trouve:=false;     // Recherche de l'etoile 'photometrie' dans la liste 'astrometrie' :
            repeat
               inc(j);
               if (Table[j].X-X)*(Table[j].X-X)+(Table[j].Y-Y)*(Table[j].Y-Y)<1 then trouve:=true;  // identité à 1 pixel près
            until (j=nb_etoile) or trouve;
         end
         else
         begin
            PhotoObjet.trouve:=true;  // Cas du fichier .csv
            j:=i;
         end;

         Table[j].trouve:=PhotoObjet.trouve;
         Table[j].EstEtalon:=PhotoObjet.trouve;

         if Table[j].trouve then    // j = index dans Table[] où il y a correspondance.
         begin
            Case Origine of
               PRISMV11  : begin
// On complète le remplissage de Table[])
               mot:=Copy(buffer,26,8); mot:=trim(mot);   // Mot 3 : FwhmX
               Val(mot,Table[j].FwhmX,code);
               mot:=Copy(buffer,35,8); mot:=trim(mot);   // Mot 4 : FwhmY
               Val(mot,Table[j].FwhmY,code);
               mot:=Copy(buffer,44,8); mot:=trim(mot);   // Mot 5 : Angle
               Val(mot,Table[j].Angle,code);
               mot:=Copy(buffer,53,6); mot:=trim(mot);   // Mot 6 : Mag
               Val(mot,Table[j].Mag,code);
               mot:=Copy(buffer,62,13); mot:=trim(mot);  // Mot 7 : flux
               Val(mot,Table[j].Flux,code);
               mot:=Copy(buffer,76,10); mot:=trim(mot);  // Mot 8 : o-c
               Val(mot,Table[j].OMC,code);
               mot:=Copy(buffer,87,7); mot:=trim(mot);   // Mot 9 : S/B
               Val(mot,Table[j].SNR,code);
            end;
               //             else pas nécessaire dans le cas du fichier .csv
            end;

            k:=0;

            Table[j].zone.nb_source:=0;

            repeat             // Reconnaissance des sources GAIA à partir des coordonnées des étoiles du CCD
                               // on enregistre toutes les sources répondant au critère
               inc(k);

               if CheckBoxMagLimite.Checked then
               case ComboBoxFiltre.ItemIndex of
               0 : MagTest:=CCDZone.Liste_source[k].magG;
               1 : MagTest:=CCDZone.Liste_source[k].magG_BP;
               2 : MagTest:=CCDZone.Liste_source[k].magG_RP;
               end
               else
                  MagTest:=CCDZone.Liste_source[k].magG;

               if (Abs(CCDZone.Liste_source[k].RA-Table[j].RA)<=FWHMmax) and { (CCDZone.Liste_source[k].RA<=Table[j].RA+FWHMmax) and  }
                  (Abs(CCDZone.Liste_source[k].DEC-Table[j].DEC)<=FWHMmax) and {and (CCDZone.Liste_source[k].DEC<=Table[j].DEC+FWHMmax) }
                  (MagTest<MagLimite)
               then
                        // Ajout de la source trouvée dans GRAPPA à la liste des sources identifiées à l'objet Table[j]
               begin
                  Inc(Table[j].zone.nb_source);
                  Nb_source:=Table[j].zone.nb_source;
                  SetLength(Table[j].zone.Liste_source,nb_source+1);
                  Table[j].zone.Liste_source[nb_source]:=CCDZone.Liste_source[k];

               end
            until (k=CCDZone.nb_source) {or Table[j].zone.overflow};

         end else Inc(cpt_non_trouve);
      end;

      if debog {and (Origine=PRISMV11) } then Writeln(fd,DateTimeToStr(now),' Nombre d''etoiles de la photometrie non correlees a l''astrometrie : ',cpt_non_trouve,'. Avec PRISM devrait être nul.');

// Recherche du premier étalon défini par le catalogue :

      if debog then Writeln(fd,DateTimeToStr(now),' Calcul des magnitudes de l''image par les flux.');

      cpt_passe:=1;
      repeat

// Méthode du superétalon :

         i:=0;
         MagEtalon_1:=99;
         repeat
            inc(i);
            with Table[i] do if (zone.nb_source>0) and EstEtalon then
            begin
               magCollapseG:=0;  magCollapseG_BP:=0; magCollapseG_RP:=0;
               for j:=1 to zone.nb_source do
               begin
                  magCollapseG   :=magCollapseG   +power10(-zone.Liste_source[j].magG/2.5);
                  magCollapseG_BP:=magCollapseG_BP+power10(-zone.Liste_source[j].magG_BP/2.5);
                  magCollapseG_RP:=magCollapseG_BP+power10(-zone.Liste_source[j].magG_RP/2.5);
               end;
               magCollapseG   :=-2.5*ln(magCollapseG   )/ln10;
               magCollapseG_BP:=-2.5*ln(magCollapseG_BP)/ln10;
               magCollapseG_RP:=-2.5*ln(magCollapseG_RP)/ln10;
               case ComboBoxFiltre.ItemIndex of
               0 : MagEtalon_1:=zone.Liste_source[1].magG;    // Filtre:=G;
               1 : MagEtalon_1:=zone.Liste_source[1].magG_BP; // Filtre:=G_BP;
               2 : MagEtalon_1:=zone.Liste_source[1].magG_RP; // Filtre:=G_RP;
               else MagEtalon_1:=zone.Liste_source[1].magG;
               end
            end;
         until MagEtalon_1<>99;


         if debog then
         begin
            Writeln(fd,DateTimeToStr(now),' Magnitude du premier étalon : ',MagEtalon_1:6:3);
            flush(fd);
         end;
         FluxTheorique:=0;
         for i:=1 to nb_etoile do
         begin
            MagEtalon_i:=0;
            if (Table[i].zone.nb_source>0) and Table[i].EstEtalon then
            for j:=1 to Table[i].zone.nb_source do
            begin
               case ComboBoxFiltre.ItemIndex of
               0 : MagEtalon_i:=Table[i].zone.Liste_source[1].magG;    // Filtre:=G;
               1 : MagEtalon_i:=Table[i].zone.Liste_source[1].magG_BP; // Filtre:=G_BP;
               2 : MagEtalon_i:=Table[i].zone.Liste_source[1].magG_RP; // Filtre:=G_RP;
               else MagEtalon_i:=Table[i].zone.Liste_source[1].magG;
               end;
               FluxTheorique:=FluxTheorique+Power10((MagEtalon_1-MagEtalon_i)/2.5);
            end;
         end;
         if debog then Writeln(fd,DateTimeToStr(now),' Flux Theorique total : ',FluxTheorique:10:3);

         MagEtalon:=MagEtalon_1-2.5*Log(FluxTheorique);
         if debog then Writeln(fd,DateTimeToStr(now),' Magnitude du super étalon : ',MagEtalon:6:3);
         if debog then flush(fd);
         FluxEtalon:=0;
         for i:=1 to nb_etoile do
         begin
            with Table[i] do if EstEtalon then FluxEtalon:=FluxEtalon+Flux;
         end;
         if debog then
         begin
            Writeln(fd,DateTimeToStr(now),' Flux réel du super étalon : ',FluxEtalon:6:3);
            flush(fd);
         end;
// Fin de la méthode par le superétalon

         n:=0;

         SommeConstante:=0;
         constanteMoyenne:=0;
         for i:=1 to nb_etoile do
         begin
             with Table[i] do if (zone.nb_source>0) and EstEtalon then
             begin
                if flux>0 then inc(n);
                Constante:=0;
                case ComboBoxFiltre.ItemIndex of
                0 : if flux>0 then Constante:=zone.Liste_source[1].magG+2.5*Log(Flux);    // Filtre:=G;
                1 : if flux>0 then Constante:=zone.Liste_source[1].magG_BP+2.5*Log(Flux); // Filtre:=G_BP;
                2 : if flux>0 then Constante:=zone.Liste_source[1].magG_RP+2.5*Log(Flux); // Filtre:=G_RP;
                else if flux>0 then Constante:=zone.Liste_source[1].magG+2.5*Log(Flux);
                end;
                if flux>0 then SommeConstante:=SommeConstante+Constante;
             end;
         end;

         if n>0 then ConstanteMoyenne:=SommeConstante/n else Writeln(fd,DateTimeToStr(now),' ATTENTION ! Nombre d''etalons nul !');
         if debog then
         begin
            Writeln(fd,DateTimeToStr(now),' Constante moyenne de point Zero photométrique : ',constanteMoyenne);
            flush(fd);
         end;

         SigmaConstante:=0;
         for i:=1 to nb_etoile do
         begin
             with Table[i] do if (zone.nb_source>0) and EstEtalon then
             begin
                sigmaConstante:=sigmaConstante+(ConstanteMoyenne-Constante)*(ConstanteMoyenne-Constante);
             end;
         end;
         SigmaConstante:=sqrt(SigmaConstante/(n-1));
         if debog then Writeln(fd,DateTimeToStr(now),' Ecart-type sur la constante photométrique de point Zero : ',Sigmaconstante);

         SommeSigmaMesure:=0;
         SommeSigmaPrism:=0;

{ Méthode du superEtalon : }
         case ComboBoxFiltre.ItemIndex of
            0 : for i:=1 to nb_etoile do
            with Table[i] do
               if (zone.nb_source>0 ) then
               begin
                  Mag_calc:=MagEtalon+2.5*Log(FluxEtalon/Flux);
                  SommeSigmaMesure:=SommeSigmaMesure+(mag_calc-zone.Liste_source[1].magG)*(mag_calc-zone.Liste_source[1].magG);
                  SommeSigmaPrism:=SommeSigmaPrism+(Mag-zone.Liste_source[1].magG)*(Mag-zone.Liste_source[1].magG);
               end;
            1:
            with Table[i] do
               if (zone.nb_source>0 ) then
               begin
                  Mag_calc:=MagEtalon+2.5*Log(FluxEtalon/Flux);
                  SommeSigmaMesure:=SommeSigmaMesure+(mag_calc-zone.Liste_source[1].magG_BP)*(mag_calc-zone.Liste_source[1].magG_BP);
                  SommeSigmaPrism:=SommeSigmaPrism+(Mag-zone.Liste_source[1].magG_BP)*(Mag-zone.Liste_source[1].magG_BP);
               end;
            2:
            with Table[i] do
               if (zone.nb_source>0 ) then
               begin
                  Mag_calc:=MagEtalon+2.5*Log(FluxEtalon/Flux);
                  SommeSigmaMesure:=SommeSigmaMesure+(mag_calc-zone.Liste_source[1].magG_RP)*(mag_calc-zone.Liste_source[1].magG_RP);
                  SommeSigmaPrism:=SommeSigmaPrism+(Mag-zone.Liste_source[1].magG_RP)*(Mag-zone.Liste_source[1].magG_RP);
               end;
         end;
{
// Méthode de la constante moyenne :

         for i:=1 to nb_etoile do
            with Table[i] do
               if (zone.nb_source>0 ) then
               begin
                  Mag_calc:=ConstanteMoyenne-2.5*Log(Flux);
                  SommeSigmaMesure:=SommeSigmaMesure+(mag_calc-zone.Liste_source[1].magG)*(mag_calc-zone.Liste_source[1].magG); // valable que pour G (à généraliser)
                  SommeSigmaPrism:=SommeSigmaPrism+(Mag-zone.Liste_source[1].magG)*(Mag-zone.Liste_source[1].magG);
               end;
}
         SigmaMesure:=sqrt(SommeSigmaMesure/(n-1));
         SigmaPrism :=sqrt(SommeSigmaPrism/(n-1));

         cpt_etalon_supprime:=0;
         case ComboBoxFiltre.ItemIndex of
            0: for i:=1 to nb_etoile do
            with Table[i] do
               if (zone.nb_source>0 ) then begin
                  if Abs(mag_calc-zone.Liste_source[1].magG)>0.3 {sigmaMesure/1.5} then begin EstEtalon:=false; inc(cpt_etalon_supprime); end;
               end;
            1: for i:=1 to nb_etoile do
            with Table[i] do
               if (zone.nb_source>0 ) then begin
                  if Abs(mag_calc-zone.Liste_source[1].magG_BP)>0.3 {sigmaMesure/1.5} then begin EstEtalon:=false; inc(cpt_etalon_supprime); end;
               end;
            2: for i:=1 to nb_etoile do
            with Table[i] do
               if (zone.nb_source>0 ) then begin
                  if Abs(mag_calc-zone.Liste_source[1].magG_RP)>0.3 {sigmaMesure/1.5} then begin EstEtalon:=false; inc(cpt_etalon_supprime); end;
               end;
         end;

         if debog then Writeln(fd,DateTimeToStr(now),' Ecart-type par recalcul des magnitudes par le flux : ',sigmaMesure:6:4,' magnitude');
         if debog then Writeln(fd,DateTimeToStr(now),' Ecart-type des écarts des magnitudes de PRISMV11   : ',sigmaPrism:6:4,' magnitude');

         if debog then Writeln(fd,DateTimeToStr(now),' Passe de calcul de magnitude N°',cpt_passe,' .... ' ,cpt_etalon_supprime,' etalon supprimés');
         inc(cpt_passe);

         flush(fd);
      until cpt_passe>1;





      Writeln(fo,'Num.: X      Y       RA (hms)       DEC (d''")       RA (d.dec)    DEC (d.dec)        Flux   magCCD   magCalc  NbID                     source_ID  RA GRAPPA    DEC GRAPPA   mag G mag BP mag RP');

      cpt_non_trouve:=0;  { maintenant on compte les non-correspondances avec GRAPPA }
      cpt_trouve:=0;
      cpt_corr_mult:=0;   { et les multiples correspondances }

      for i:=1 to nb_etoile do
      begin
         With Table[i] do
           if zone.nb_source>0 then
           begin
              inc(cpt_trouve);
              if zone.nb_source=1 then
                 Writeln(fo,i:5,' ',X:6:1,' ',Y:6:1,'  ',RAhms,'  ',DEChms,' ',RA:12:8,' ',DEC:12:8,'  ',Flux:10:0,'   ',mag:6:3,'   ',mag_calc:6:3,'   ',
                         zone.nb_source:2,zone.Liste_source[1].Source_ID:32,' ',zone.Liste_source[1].RA:12:8,' ',zone.Liste_source[1].DEC:12:8,' ',
                         zone.Liste_source[1].magG:6:3,' ',zone.Liste_source[1].magG_BP:6:3,' ',zone.Liste_source[1].magG_RP:6:3)
              else
              begin
                 inc(cpt_corr_mult);
                 Write(fo,i:5,' ',X:6:1,' ',Y:6:1,'  ',RAhms,'  ',DEChms,' ',RA:12:8,' ',DEC:12:8,'  ',Flux:10:0,'   ',mag:6:3,'   ',mag_calc:6:3,'   ',
                       zone.nb_source:2);
                 if Not(DoCollapse) and Not(DoBrightest) then
                 begin
                    Writeln(fo,zone.Liste_source[1].Source_ID:32,' ',zone.Liste_source[1].RA:12:8,' ',zone.Liste_source[1].DEC:12:8,' ',
                         zone.Liste_source[1].magG:6:3,' ',zone.Liste_source[1].magG_BP:6:3,' ',zone.Liste_source[1].magG_RP:6:3);

                    For j:=2 to zone.nb_source do
                    Writeln(fo,i:5,'                                                                                               ',
                            zone.nb_source:2,zone.Liste_source[j].Source_ID:32,' ',zone.Liste_source[j].RA:12:8,' ',zone.Liste_source[j].DEC:12:8,' ',
                            zone.Liste_source[j].magG:6:3,' ',zone.Liste_source[j].magG_BP:6:3,' ',zone.Liste_source[j].magG_RP:6:3);
                 end else
                 if DoBrightest then
                 begin
                    iMax:=1; magMaxBright:=100;
                    for j:=1 to zone.nb_source do

                       if zone.Liste_source[j].magG<magMaxBright then begin
                       begin
                          iMax:=j;
                          magMaxBright:=zone.Liste_source[j].magG
                       end;
                    end;
                    Writeln(fo,zone.Liste_source[iMax].Source_ID:32,' ',zone.Liste_source[iMax].RA:12:8,' ',zone.Liste_source[iMax].DEC:12:8,' ',
                               zone.Liste_source[iMax].magG:6:3,' ',zone.Liste_source[iMax].magG_BP:6:3,' ',zone.Liste_source[iMax].magG_RP:6:3);
                 end
                 else
                 begin  // Cas DoCollapse
                    magCollapseG:=0;  magCollapseG_BP:=0; magCollapseG_RP:=0;
                    for j:=1 to zone.nb_source do
                    begin
                       magCollapseG   :=magCollapseG   +power10(-zone.Liste_source[j].magG/2.5);
                       magCollapseG_BP:=magCollapseG_BP+power10(-zone.Liste_source[j].magG_BP/2.5);
                       magCollapseG_RP:=magCollapseG_BP+power10(-zone.Liste_source[j].magG_RP/2.5);
                    end;
                    magCollapseG   :=-2.5*ln(magCollapseG   )/ln10;
                    magCollapseG_BP:=-2.5*ln(magCollapseG_BP)/ln10;
                    magCollapseG_RP:=-2.5*ln(magCollapseG_RP)/ln10;
                    Writeln(fo,zone.Liste_source[1].Source_ID:32,' ',zone.Liste_source[1].RA:12:8,' ',zone.Liste_source[1].DEC:12:8,' ',
                               magCollapseG:6:3,' ',magCollapseG_BP:6:3,' ',magCollapseG_RP:6:3);
                 end;
              end;
           end
           else begin
              Writeln(fo,i:5,' ',X:6:1,' ',Y:6:1,'  ',RAhms,'  ',DEChms,' ',RA:12:8,' ',DEC:12:8,'  ',Flux:10:0,'   ',mag:6:3);
              Inc(cpt_non_trouve);
           end;
      end;

      if debog then
      begin
         Writeln(fd,DateTimeToStr(now),' ',cpt_non_trouve,' etoiles sans correspondance dans GRAPPA.');
         Write  (fd,DateTimeToStr(now),' ',cpt_trouve,' etoiles avec correspondance dans GRAPPA dont ');
         Writeln(fd,cpt_corr_mult,' etoile(s) avec correspondances mutliples dans GRAPPA.');
      end;

      // On libère la mémoire :

      for i:=1 to nb_etoile do
          With Table[i] do SetLength(Table[i].zone.Liste_source,0);
      SetLength(CCDZone.Liste_source,0);
      if debog then Writeln(fd,DateTimeToStr(now),' Mémoire libérée.');
   end
   else
      Writeln(fo,'Aucune etoile trouvee dans la zone de recherche dans GRAPPA.');

   SetLength(Table,0);

   CloseFile(f_astro);
   if debog then Closefile(fd);

   CloseFile(fo);
   MemoOut.Lines.LoadFromFile(nf_out);     // On récupère les résultats depuis le fichier de sortie
   DeleteFile(nf_out);

   SaveDialog1.FileName:=nf_out;
   SaveDialog1.Title:='Nom du fichier résultat ?';
   repeat
      DoSauve:=SaveDialog1.Execute;
   until DoSauve;
   If DoSauve then
   begin
      if FileExists(SaveDialog1.FileName) and (SaveDialog1.FileName<>nf_out) then
      begin
         With Application do i:=MessageBox('Ecraser le fichier existant ?','Etes vous sûr ?',MB_YESNO);
         if i=ID_YES then
         begin
            DeleteFile(SaveDialog1.FileName);
            MemoOut.Lines.SaveToFile(SaveDialog1.FileName);
         end;
      end
      else
         MemoOut.Lines.SaveToFile(SaveDialog1.FileName);
   end;


   ButtonAnalyser.Caption:='Analyser';

end;


procedure TForm1.ButtonBaseClick(Sender: TObject);
begin
    if SelectDirectoryDialog1.Execute then
       if DirectoryExists(SelectDirectoryDialog1.FileName) then
       begin
          base:=SelectDirectoryDialog1.Filename;
          LabeledEditBase.Text:=base;
          LabeledEditBase.Font.Color:=clGreen;

       end;
end;

procedure TForm1.ButtonAideClick(Sender: TObject);
begin
   Form2.show; Form2.Visible:=true;
end;

procedure TForm1.CheckBoxXMaxChange(Sender: TObject);
begin
  if CheckBoxCollapse.checked and CheckBoxXMax.Checked then CheckBoxCollapse.checked:=false;;
end;

procedure TForm1.CheckBoxMagLimiteChange(Sender: TObject);
begin
   if CheckBoxMaglimite.Checked=true then
   begin
      EditMaglimite.Visible:=true;
      EditMaglimite.Enabled:=true;
      Label1.Visible:=true;
      ComboBoxFiltre.Visible:=true;
   end
   else
   begin
      EditMaglimite.Visible:=false;
      EditMaglimite.Enabled:=false;
      EditMagLimite.Text:='99';
      Label1.Visible:=True;
      ComboBoxFiltre.Visible:=True;
   end;
end;

procedure TForm1.CheckBoxErreurChange(Sender: TObject);
begin
   if CheckboxErreur.Checked=true then
      With EditErreur do
      begin
         Visible:=true;
         Enabled:=true;
      end
   else
   begin
      With EditErreur do
      begin
         Visible:=false;
         Enabled:=false;
      end;
   end;
end;

procedure TForm1.EditErreurChange(Sender: TObject);
var
   code   : integer;
   Erreur : real;
begin

   Val(EditErreur.Text,erreur,code);

   if (code<>0) or (erreur<=0) then
      EditErreur.Font.Color:=clRed
   else
      EditErreur.Font.Color:=clDefault;
end;

procedure TForm1.EditMagLimiteChange(Sender: TObject);
var
   code : integer;
   Limite : real;
begin

   Val(EditMagLimite.Text,limite,code);

   if code<>0 then
      EditMagLimite.Font.Color:=clRed
   else
      if limite>25 then EditMagLimite.Font.Color:=clDefault else EditMagLimite.Font.Color:=ClGreen;
end;

procedure TForm1.LabeledEditBaseChange(Sender: TObject);
begin

end;

procedure TForm1.RadioButtonPRISM11Change(Sender: TObject);
begin

end;



procedure TForm1.FormCreate(Sender: TObject);
var
   fi : textfile;
   donnees : string;
begin
   ok_astro:=false;
   ok_photo:=true;
   ok_erreur:=false;
   ok_collapse:=false;
   ok_xmax:=false;
   nf_astro:='';
   nf_photo:='';
   base:='';
   ButtonAnalyser.Enabled:=false;
   LabeledEditDonnees.Text:='Pas de données';
   donnees:='';

   if FileExists(nf_ini) then
   begin
      AssignFile(fi,nf_ini);
      Reset(fi);
      Readln(fi);
      Readln(fi,base);
      Readln(fi);
      Readln(fi,donnees);
      CloseFile(fi);
   end;
   LabeledEditDonnees.Text:=donnees;

   if FileExists(donnees) then
   begin
      ok_astro:=true;
      nf_astro:=donnees;
      LabeledEditDonnees.Text:=nf_astro;
      LabeledEditDonnees.Font.Color:=clDefault;
   end
   else
   begin
      ok_astro:=false;
      LabeledEditDonnees.Font.Color:=clRed;
   end;

   if DirectoryExists(Base) then
   begin
      LabeledEditBase.Text:=base;
      LabeledEditBase.Font.Color:=clGreen;
   end else
   begin
      LabeledEditBase.Text:='';
      Base:='';
   end;

   ButtonAnalyser.Enabled:=ok_astro and (LabeledEditBase.Text<>'');

end;

procedure TForm1.FormDestroy(Sender: TObject);
var
   fi : textfile;
begin
   AssignFile(fi,nf_ini);
   Rewrite(fi);
   Writeln(fi,'// Répertoire de la base GRAPPA :');
   Writeln(fi,LabeledEditBase.Text);
   Writeln(fi,'// Nom du fichier de données : ');
   Writeln(fi,LabeledEditDonnees.Text);
   CloseFile(fi);
end;

end.

